@@ -112,15 +112,19 @@ sub parse_on_start($$%)
 
   $tag = lc($tag);
 
-  if($tag eq "function")
+  if($tag eq "function" or $tag eq "signal" or $tag eq "enum")
   {
     if(defined $DocsParser::objCurrentFunction)
     {
-      $objParser->xpcroak("\nClose a function tag before you open another one.");
+      $objParser->xpcroak(
+        "\nClose a function, signal or enum tag before you open another one.");
     }
 
     my $functionName = $attr{name};
 
+    # Change signal name from Class::a-signal-name to Class::a_signal_name.
+    $functionName =~ s/-/_/g if($tag eq "signal");
+
     #Reuse existing Function, if it exists:
     #(For instance, if this is the override parse)
     $DocsParser::objCurrentFunction = $DocsParser::hasharrayFunctions{$functionName};
@@ -191,7 +195,7 @@ sub parse_on_end($$)
 
   $tag = lc($tag);
 
-  if($tag eq "function")
+  if($tag eq "function" or $tag eq "signal" or $tag eq "enum")
   {
     # Store the Function structure in the array:
     my $functionName = $$DocsParser::objCurrentFunction{name};
@@ -229,11 +233,92 @@ sub parse_on_cdata($$)
   }
 }
 
-
-# $strCommentBlock lookup_documentation($strFunctionName)
-sub lookup_documentation($$)
+sub lookup_enum_documentation($$$)
 {
-  my ($functionName, $deprecation_docs) = @_;
+  my ($c_enum_name, $cpp_enum_name, $ref_flags) = @_;
+  
+  my @subst_in  = [];
+  my @subst_out = [];
+ 
+ # Get the substitutions.
+  foreach(@$ref_flags)
+  {
+    if(/^\s*s#([^#]+)#([^#]*)#\s*$/)
+    {
+      push(@subst_in,  $1);
+      push(@subst_out, $2);
+    }
+  }
+
+  my $objFunction = $DocsParser::hasharrayFunctions{$c_enum_name};
+  if(!$objFunction)
+  {
+    #print "DocsParser.pm: Warning: enum not found: $enum_name\n";
+    return ""
+  }
+
+  my $docs = "";
+
+  my @param_names = @{$$objFunction{param_names}};
+  my $param_descriptions = \$$objFunction{param_descriptions};
+
+  # Append the param docs first so that the enum description can come last and
+  # the possible flag docs that the m4 _ENUM() macro appends goes in the right
+  # place.
+  foreach my $param (@param_names)
+  {
+    my $desc = $$param_descriptions->{$param};
+
+    # Remove the initial prefix in the name of the enum constant. Would be something like GTK_.
+    $param =~ s/\b[A-Z]+_//;
+
+    # Now apply custom substitutions.
+    for(my $i = 0; $i < scalar(@subst_in); ++$i)
+    {
+      $param =~ s/${subst_in[$i]}/${subst_out[$i]}/;
+      $desc  =~ s/${subst_in[$i]}/${subst_out[$i]}/;
+    }
+
+    # Skip this element, if its name has been deleted.
+    next if($param eq "");
+
+    $param =~ s/([a-zA-Z0-9]*(_[a-zA-Z0-9]+)*)_?/$1/g;
+    if(length($desc) > 0)
+    {
+      $desc =~ s/\n/ /g;
+      $desc =~ s/ $//;
+      $desc =~ s/^\s+//; # Chop off leading whitespace
+      $desc .= '.' unless($desc =~ /(?:^|\.)$/);
+      $docs .= "\@var $cpp_enum_name ${param}\n \u${desc}\n\n"; # \u = Convert next char to uppercase
+    }
+  }
+
+  # Append the enum description docs.
+  $docs .= "\@enum $cpp_enum_name\n"; 
+  $docs .= $$objFunction{description};
+
+  DocsParser::convert_docs_to_cpp($objFunction, \$docs);
+  DocsParser::add_m4_quotes(\$docs);
+
+  # Escape the space after "i.e." or "e.g." in the brief description.
+  $docs =~ s/^([^.]*\b(?:i\.e\.|e\.g\.))\s/$1\\ /;
+  
+  remove_example_code($c_enum_name, \$docs);
+
+  # Convert to Doxygen-style comment.
+  $docs =~ s/\n/\n \* /g;
+  $docs =  "\/\*\* " . $docs;
+
+  return $docs;
+}
+
+# $strCommentBlock lookup_documentation($strFunctionName, $deprecation_docs, $objCppfunc)
+# The final objCppfunc parameter is optional.  If passed, it is used to
+# decide if the final C parameter should be omitted if the C++ method
+# has a slot parameter.
+sub lookup_documentation($$;$)
+{
+  my ($functionName, $deprecation_docs, $objCppfunc) = @_;
 
   my $objFunction = $DocsParser::hasharrayFunctions{$functionName};
   if(!$objFunction)
@@ -246,7 +331,7 @@ sub lookup_documentation($$)
 
   if(length($text) eq 0)
   {
-    print "DocsParser.pm: Warning: No C docs for function: \"$functionName\"\n";
+    print "DocsParser.pm: Warning: No C docs for: \"$functionName\"\n";
   }
 
   DocsParser::convert_docs_to_cpp($objFunction, \$text);
@@ -257,13 +342,15 @@ sub lookup_documentation($$)
     $text .= "\n\@deprecated $deprecation_docs";
   }
 
-  DocsParser::append_parameter_docs($objFunction, \$text);
+  DocsParser::append_parameter_docs($objFunction, \$text, $objCppfunc);
   DocsParser::append_return_docs($objFunction, \$text);
   DocsParser::add_m4_quotes(\$text);
 
   # Escape the space after "i.e." or "e.g." in the brief description.
   $text =~ s/^([^.]*\b(?:i\.e\.|e\.g\.))\s/$1\\ /;
 
+  remove_example_code($functionName, \$text);
+
   # Convert to Doxygen-style comment.
   $text =~ s/\n/\n${DocsParser::commentMiddleStart}/g;
   $text =  $DocsParser::commentStart . $text;
@@ -272,6 +359,23 @@ sub lookup_documentation($$)
   return $text;
 }
 
+# void remove_example_code($obj_name, \$text)
+# Removes example code from the text of docs (passed by reference).
+sub remove_example_code($$)
+{
+  my ($obj_name, $text) = @_;
+
+  # Remove C example code.
+  my $example_removals =
+    ($$text =~ s"<informalexample>.*?</informalexample>"[C example ellipted]"sg);
+  $example_removals +=
+    ($$text =~ s"<programlisting>.*?</programlisting>"\n[C example ellipted]"sg);
+  $example_removals += ($$text =~ s"\|\[.*?]\|"\n[C example ellipted]"sg);
+
+  print STDERR "gmmproc: $main::source: $obj_name: Example code discarded.\n"
+    if ($example_removals);
+}
+
 sub add_m4_quotes($)
 {
   my ($text) = @_;
@@ -287,9 +391,12 @@ sub add_m4_quotes($)
   $$text = "`" . $$text . "'";
 }
 
-sub append_parameter_docs($$)
+# The final objCppfunc is optional.  If passed, it is used to determine
+# if the final C parameter should be omitted if the C++ method has a
+# slot parameter.
+sub append_parameter_docs($$;$)
 {
-  my ($obj_function, $text) = @_;
+  my ($obj_function, $text, $objCppfunc) = @_;
 
   my @param_names = @{$$obj_function{param_names}};
   my $param_descriptions = \$$obj_function{param_descriptions};
@@ -301,12 +408,27 @@ sub append_parameter_docs($$)
   shift(@param_names) if(($defs_method && $$defs_method{class} ne "") ||
                          ($$obj_function{mapped_class} ne ""));
 
+  # Also skip first param if this is a signal.
+  shift(@param_names) if ($$obj_function{name} =~ /\w+::/);
+
+  # Skip the last param if there is a slot because it would be a
+  # gpointer user_data parameter.
+  pop(@param_names) if (defined($objCppfunc) && $$objCppfunc{slot_name});
+
   foreach my $param (@param_names)
   {
     if ($param ne "error" ) #We wrap GErrors as exceptions, so ignore these.
     {
       my $desc = $$param_descriptions->{$param};
 
+      # Deal with callback parameters converting the docs to a slot
+      # compatible format.
+      if ($param eq "callback")
+      {
+        $param = "slot";
+        $$text =~ s/\@a callback/\@a slot/g;
+      }
+
       $param =~ s/([a-zA-Z0-9]*(_[a-zA-Z0-9]+)*)_?/$1/g;
       DocsParser::convert_docs_to_cpp($obj_function, \$desc);
       if(length($desc) > 0)
@@ -359,52 +481,90 @@ sub convert_tags_to_doxygen($)
   for($$text)
   {
     # Replace format tags.
-    s"<(/?)emphasis>"<$1em>"g;
-    s"<(/?)literal>"<$1tt>"g;
-    s"<(/?)constant>"<$1tt>"g;
-    s"<(/?)function>"<$1tt>"g;
+    s"<(/?)(?:emphasis|replaceable)>"<$1em>"g;
+    s"<(/?)(?:constant|envar|filename|function|guimenuitem|literal|option|structfield|varname)>"<$1tt>"g;
 
     # Some argument names are suffixed by "_" -- strip this.
     # gtk-doc uses @thearg, but doxygen uses @a thearg.
     s" ?\@([a-zA-Z0-9]*(_[a-zA-Z0-9]+)*)_?\b" \@a $1"g;
 
-    # Don't convert doxygen's @throws and @param, so these can be used in the
-    # docs_override.xml:
-    s" \@a (throws|param)" \@$1"g;
-    s"^Note ?\d?: "\@note "mg;
+    # Don't convert Doxygen's $throw, @throws and @param, so these can be used
+    # in the docs_override.xml.
+    # Also don't convert @enum and @var which are used for enum documentation.
+    s" \@a (throws?|param|enum|var)\b" \@$1"g;
 
+    s"^Note ?\d?: "\@note "mg;
     s"</?programlisting>""g;
-    s"<informalexample>|\|\["\@code"g;
-    s"</informalexample>|]\|"\@endcode"g;
     s"<!>""g;
 
     # Remove all link tags.
     s"</?u?link[^&]*?>""g;
 
-    # Remove all para tags (from tmpl sgml files).
-    s"</?para>""g;
+    # Remove all para tags and simpara tags (simple paragraph).
+    s"</?(sim)?para>""g;
 
-    # Use our Doxygen @newin alias:
-    s/\bSince:\s*(\d+)\.(\d+)\b/\@newin{$1,$2}/g;
+    # Convert <simplelist>, <itemizedlist> and <variablelist> to something that
+    # Doxygen understands.
+    s"<simplelist>\n?(.*?)</simplelist>\n?"&DocsParser::convert_simplelist($1)"esg;
+    s"<itemizedlist>\n?(.*?)</itemizedlist>\n?"&DocsParser::convert_itemizedlist($1)"esg;
+    s"<variablelist>\n?(.*?)</variablelist>\n?"&DocsParser::convert_variablelist($1)"esg;
+
+    # Use our Doxygen @newin alias.
+    # If Since is not followed by a colon, substitute @newin only if it's
+    # in a sentence of its own at the end of the string. 
+    s/\bSince:\s*(\d+)\.(\d+)\.(\d+)\b\.?/\@newin{$1,$2,$3}/g;
+    s/\bSince:\s*(\d+)\.(\d+)\b\.?/\@newin{$1,$2}/g;
+    s/(\.\s+)Since\s+(\d+)\.(\d+)\.(\d+)\.?$/$1\@newin{$2,$3,$4}/;
+    s/(\.\s+)Since\s+(\d+)\.(\d+)\.?$/$1\@newin{$2,$3}/;
 
     s"\b->\b"->"g;
 
     # Doxygen is too dumb to handle &mdash;
     s"&mdash;" \@htmlonly&mdash;\@endhtmlonly "g;
 
-    s"\%?FALSE\b"<tt>false</tt>"g;
-    s"\%?TRUE\b"<tt>true</tt>"g;
-    s"\%?NULL\b"<tt>0</tt>"g;
+    s"\%?\bFALSE\b"<tt>false</tt>"g;
+    s"\%?\bTRUE\b"<tt>true</tt>"g;
+    s"\%?\bNULL\b"<tt>0</tt>"g;
 
     s"#?\bgboolean\b"<tt>bool</tt>"g;
     s"#?\bg(int|short|long)\b"<tt>$1</tt>"g;
     s"#?\bgu(int|short|long)\b"<tt>unsigned $1</tt>"g;
 
-    # For Gtk::TextIter.
-    s"(\\[rn])\b"<tt>\\$1</tt>"g;
+    # Escape all backslashes, except in \throw, \throws and \param, which can
+    # be Doxygen commands in the docs_override.xml.
+    s"\\"\\\\"g;
+    s"\\\\(throws?|param)\b"\\$1"g
   }
 }
 
+# Convert <simplelist> tags to a list of newline-separated elements.
+sub convert_simplelist($)
+{
+  my ($text) = @_;
+
+  $text =~ s"<member>(.*?)(\n?)</member>(\n?)"$1<br>\n"sg;
+  return "<br>\n" . $text . "<br>\n";
+}
+
+# Convert <itemizedlist> tags to Doxygen format.
+sub convert_itemizedlist($)
+{
+  my ($text) = @_;
+
+  $text =~ s"<listitem>(.*?)(\n?)</listitem>(\n?)"- $1\n"sg;
+  return $text;
+}
+
+# Convert <variablelist> tags to an HTML definition list.
+sub convert_variablelist($)
+{
+  my ($text) = @_;
+
+  $text =~ s"</?varlistentry>\n?""g;
+  $text =~ s"<(/?)term>"<$1dt>"g;
+  $text =~ s"<(/?)listitem>"<$1dd>"g;
+  return "<dl>\n" . $text . "</dl>\n";
+}
 
 sub substitute_identifiers($$)
 {
@@ -414,6 +574,26 @@ sub substitute_identifiers($$)
   {
     # TODO: handle more than one namespace
 
+    # Convert property names to C++.
+    # The standard (and correct) gtk-doc way of referring to properties.
+    s/(#[A-Z]\w+):([a-z\d-]+)/my $name = "$1::property_$2()"; $name =~ s"-"_"g; "$name";/ge;
+    # This is an incorrect format but widely used so correctly treat as a
+    # property.
+    s/(\s)::([a-z\d-]+)(\s+property)/my $name = "$1property_$2()$3"; $name =~ s"-"_"g; "$name";/ge;
+    # This one catches properties written in the gtk-doc block as for example
+    # '#GtkActivatable::related-action property'.  The correct way to write it
+    # would be 'GtkActivatable:related-action' (with a single colon and not
+    # two because the double colons are specifically for signals -- see the
+    # gtk-doc docs:
+    # http://developer.gnome.org/gtk-doc-manual/unstable/documenting_symbols.html.en)
+    # but a few are written with the double colon in the gtk+ docs so this
+    # protects against those errors.
+    s/([A-Z]\w+)::([a-z\d-]+)(\s+property)/my $name = "$1::property_$2()$3"; $name =~ s"-"_"g; "$name";/ge;
+
+    # Convert signal names to C++.
+    s/(^|\s)::([a-z\d-]+)(\(\))*([^:\w]|$)/my $name = "$1signal_$2()$4"; $name =~ s"-"_"g; "$name";/ge;
+    s/(#[A-Z]\w+)::([a-z\d-]+)(\(\))*([^:\w]|$)/my $name = "$1::signal_$2()$4"; $name =~ s"-"_"g; "$name";/ge;
+
     s/[#%]([A-Z][a-z]*)([A-Z][A-Za-z]+)\b/$1::$2/g; # type names
 
     s/[#%]([A-Z])([A-Z]*)_([A-Z\d_]+)\b/$1\L$2\E::$3/g; # enum values
@@ -421,7 +601,11 @@ sub substitute_identifiers($$)
     # Undo wrong substitutions.
     s/\bHas::/HAS_/g;
     s/\bNo::/NO_/g;
-    s/\bG:://g; #Rename G::Something to Something. Doesn't seem to work. murrayc.
+    s/\bO::/O_/g;
+    s/\bG:://g; #Rename G::Something to Something.
+
+    # Substitute callback types to slot types.
+    s/(\b\w+)Callback/Slot$1/g;
 
     # Replace C function names with C++ counterparts.
     s/\b([a-z]+_[a-z][a-z\d_]+) ?\(\)/&DocsParser::substitute_function($doc_func, $1)/eg;
